<?php

declare(strict_types=1);

namespace Drupal\Tests\elasticsearch_connector\Unit\SearchAPI\Query;

use Drupal\Tests\UnitTestCase;
use Drupal\elasticsearch_connector\SearchAPI\Query\FacetParamBuilder;
use Drupal\search_api\Query\QueryInterface;
use Prophecy\PhpUnit\ProphecyTrait;
use Psr\Log\LoggerInterface;

/**
 * Tests the facet param builder.
 *
 * @coversDefaultClass \Drupal\elasticsearch_connector\SearchAPI\Query\FacetParamBuilder
 * @group elasticsearch_connector
 */
class FacetParamBuilderTest extends UnitTestCase {

  use ProphecyTrait;

  /**
   * @covers ::buildFacetParams
   */
  public function testBuildFacetParams() {
    $logger = $this->prophesize(LoggerInterface::class);
    $builder = new FacetParamBuilder($logger->reveal());

    $query = $this->prophesize(QueryInterface::class);
    $query->getOption('search_api_facets', [])
      ->willReturn([
        'facet1' => [
          'field' => 'field1',
          'operator' => 'and',
        ],
        'facet2' => [
          'field' => 'field1',
          'operator' => 'or',
        ],
      ]);

    $indexFields = [
      'field1' => [],
      'field2' => [],
    ];

    $facetFilters = [
      "facet2" => "filter for facet2",
    ];

    $aggs = $builder->buildFacetParams($query->reveal(), $indexFields, $facetFilters);

    $expected = [
      'facet1_filtered' => [
        'filter' => "filter for facet2",
        'aggs' => [
          'facet1' =>
          ['terms' => ['field' => 'field1', 'size' => 10]],
        ],
      ],
      'facet2' => ['terms' => ['field' => 'field1', 'size' => 10]],
    ];

    $this->assertNotEmpty($aggs);
    $this->assertEquals($expected, $aggs);
  }

  /**
   * @covers ::buildFacetParams
   */
  public function testBuildFacetParamsAnd() {
    $logger = $this->prophesize(LoggerInterface::class);
    $builder = new FacetParamBuilder($logger->reveal());

    $query = $this->prophesize(QueryInterface::class);
    $query->getOption('search_api_facets', [])
      ->willReturn([
        'facet1' => [
          'field' => 'field1',
          'operator' => 'and',
        ],
        'facet2' => [
          'field' => 'field1',
          'operator' => 'and',
        ],
      ]);

    $indexFields = [
      'field1' => [],
      'field2' => [],
    ];

    $aggs = $builder->buildFacetParams($query->reveal(), $indexFields);

    $expected = [
      'facet1' => ['terms' => ['field' => 'field1', 'size' => 10]],
      'facet2' => ['terms' => ['field' => 'field1', 'size' => 10]],
    ];

    $this->assertNotEmpty($aggs);
    $this->assertEquals($expected, $aggs);
  }

  /**
   * @covers ::buildTermBucketAgg
   */
  public function testBuildTermBucketAgg() {
    $logger = $this->prophesize(LoggerInterface::class);
    $builder = new FacetParamBuilder($logger->reveal());

    $reflection = new \ReflectionClass($builder);
    $method = $reflection->getMethod('buildTermBucketAgg');

    // Limit is set to 5, size = 5.
    $facet_with_limit = [
      'field' => 'field_tags',
      'limit' => 5,
      'operator' => 'or',
    ];
    $agg_with_limit = $method->invokeArgs($builder, ['facet_limited', $facet_with_limit, []]);
    $this->assertEquals(5, $agg_with_limit['facet_limited']['terms']['size']);

    // Limit is set to 0 (No limit), size = 10000.
    $facet_with_zero_limit = [
      'field' => 'field_tags',
      'limit' => 0,
      'operator' => 'or',
    ];
    $agg_with_zero = $method->invokeArgs($builder, ['facet_unlimited', $facet_with_zero_limit, []]);
    $this->assertEquals(10000, $agg_with_zero['facet_unlimited']['terms']['size']);

    // Limit not set at all, default size (10) should be used.
    $facet_without_limit = [
      'field' => 'field_tags',
      'operator' => 'or',
    ];
    $agg_without_limit = $method->invokeArgs($builder, ['facet_default', $facet_without_limit, []]);
    $this->assertEquals(10, $agg_without_limit['facet_default']['terms']['size']);
  }

}
